"""
This module contains utilities for extracting decision rules and counterfactuals from a DecisionTreeClassifier.
"""
from sklearn.tree import DecisionTreeClassifier


def decision_rule(x, tree: DecisionTreeClassifier):
    """
    Extract the decision rule for the classification of x.
    :param x: instance to classify.
    :param tree: classifier.
    :return:
    """


def counterfactuals():
    pass



"""
This module contains utility functions for defining distance functions to be used by LORE's fitness function
"""
from abc import ABC, abstractmethod
import numpy as np
from numba import njit


class AbstractDistance(ABC):
    @abstractmethod
    def __call__(self, x, y, *args, **kwargs):
        pass


class EuclideanDistance(AbstractDistance):
    """
    Computes the Euclidean Distance between two vectors.
    """

    def __call__(self, x: np.array, y: np.array, *args, **kwargs):
        diff = x - y
        return np.linalg.norm(diff)


class NormalizedEuclideanDistance(AbstractDistance):
    """
    Normalized the vectors and computes the Euclidean Distance between them.
    """

    def __call__(self, x: np.array, y: np.array, *args, **kwargs):
        diff = normalize_vector(x - y)
        return np.linalg.norm(diff)


class SimpleMatchDistance(AbstractDistance):
    """

    """
    def __call__(self, x, y, *args, **kwargs):
        dist = 0
        for (a, b) in zip(x, y):
            if a != b:
                dist += 1
        return dist


class MixedAttributesDistance(AbstractDistance):
    """
    Computes the distance between two instances when these are composed of both categorical and continuous attributes.


    """
    def __init__(self, sep_index: int, **kwargs):
        self.sep_index = sep_index
        self.categ_dist = SimpleMatchDistance()
        self.neucl_dist = NormalizedEuclideanDistance()

    def __call__(self, x, y, *args, **kwargs):
        s = self.sep_index
        categorical = x[:s], y[:s]
        continuous = x[s:], y[:s]
        nelem = len(x)
        return (self.sep_index / nelem) * self.categ_dist(*categorical) + (
                (nelem - self.sep_index) / nelem) * self.neucl_dist(
            *continuous)


# utilities
def normalize_vector(x: np.array):
    norm = np.dot(np.sqrt(x))
    if norm:
        return x / norm
    else:
        return norm


@njit
def numba_simple_match(x, y):
    dist = 0
    for (a, b) in zip(x, y):
        if a != b:
            dist += 1
    return dist



from numba import njit
import numpy as np

from pylore.blackbox import AbstractBlackBoxWrapper
from pylore.distances import AbstractDistance
from sklearn import tree

from pylore.genetics import genetic_neighborhood


class LORE:
    """

    """

    def __init__(self, bb: AbstractBlackBoxWrapper, neighbors: int, distance: AbstractDistance, **kwargs):
        """

        :param bb:
        :param neighbors:
        :param distance:
        :keyword generations (int, default 10): number of generations for the genetic neighborhood construction.
        :keyword crossover_prob:
        :keyword mutation_prob:
        :keyword random_state (int): Controls the randomness of the estimator.

        """
        self.bb = bb
        self.neighbors = neighbors
        self.distance = distance

        # default values from the paper
        self.generations = kwargs.get('generations', 10)
        self.crossover_prob = kwargs.get('crossover_prob', 0.5)
        self.mutation_prob = kwargs.get('mutation_prob', 0.2)

        # instantiate the classifier
        random_state = kwargs.get('random_state')
        self.clf = tree.DecisionTreeClassifier(random_state=random_state)

    def __call__(self, x, *args, **kwargs):
        """

        :param x: instance to explain
        :param args:
        :param kwargs:
        :return:
        """



        pass





"""
This module contains the implementation of the function that creates a synthetic neighborhood of an instance using a genetic algorithm.
"""
from abc import ABC, abstractmethod

from pylore.blackbox import AbstractBlackBoxWrapper
from typing import Callable
import numpy as np
from numba import njit

from pylore.distances import AbstractDistance


class AbstractMutator(ABC):

    @abstractmethod
    def fit(self, data):
        """
        Extract the empirical distribution from the data in order to be able to mutate some attributes.
        :param data:
        :return:
        """
        pass

    @abstractmethod
    def __call__(self, x, *args, **kwargs):
        pass


class DefaultMutator(AbstractMutator):

    def __init__(self):
        # separator
        pass

    def fit(self, data):
        pass

    def __call__(self, x, *args, **kwargs):
        pass


class AbstractCrossoverer(ABC):
    @abstractmethod
    def __call__(self, *args, **kwargs):
        p1 = ...
        p2 = ...
        features = np.random.randint(0, p1.shape[0], size=(2,))


class TwoPointCrossoverer(AbstractCrossoverer):
    """

    """

    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        # select two items at random
        p1, p1 = ...
        # swap two random features
        idx = np.random.randint(0, p1.shape[0], size=(2,))
        p1[idx], p2[idx] = p2[idx], p1[idx]
        pass



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='pylore',
    url='https://github.com/gerzin/pylore',
    author='Gerardo Zinno',
    author_email='zinno.gerardo.95@gmail.com',
    version='0.0.2',
    description='LOcal Rule-based Explanations',
    packages=['pylore'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "numpy",
        "numba",
        "scikit-learn"
    ],
    extras_require={
        "dev": [
            "pytest > 3.7",
            "twine",
        ]
    }
)
